use std::str::FromStr;
use std::sync::Arc;

use cidr::IpInet;
use serde::{Deserialize, Serialize};
use tokio::sync::Mutex;

type WgLink = String;
type WgPeer = String;

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct EventInfo {
    pub link: WgLink,
    pub cidr: IpInet,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum Event {
    Add(EventInfo),
    Del(EventInfo),
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct EventParseError<T>(T);

impl FromStr for Event {
    type Err = EventParseError<String>;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let re = regex::Regex::new(r"^(Deleted )?\d+: ([a-z0-9]+)\s+inet ((\d+\.?){4}(/\d+)?).*\n?$")
            .unwrap();

        if let Some(captures) = re.captures(s) {
            let is_add = captures.get(1).is_none();
            let _is_delete = captures.get(1).is_some();
            let link: WgLink = captures.get(2).unwrap().as_str().into();
            let cidr: IpInet = IpInet::from_str(captures.get(3).unwrap().as_str()).unwrap();
            if is_add {
                Ok(Event::Add(EventInfo { link, cidr }))
            } else {
                Ok(Event::Del(EventInfo { link, cidr }))
            }
        } else {
            Err(Self::Err {
                0: format!("'{}' couldn't be parsed as event", s),
            })
        }
    }
}

impl Event {
    pub fn inner(&self) -> &EventInfo {
        match self {
            Self::Add(e) => e,
            Self::Del(e) => e,
        }
    }
}

type EventStore = Arc<Mutex<Vec<Event>>>;
