use clap::Parser;
use helpers::command_output;
use std::{str::FromStr, collections::HashMap};
use tokio::io::{self, AsyncBufReadExt, BufReader};
use warp::Filter;
use wg::Peer;

mod event;
mod helpers;
mod model;
mod wg;

type WgLink = String;
type WgPeer = String;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct CliArguments {
    #[arg(short, long)]
    addr: std::net::SocketAddr,
    #[arg(short, long)]
    link: WgLink,
}

#[derive(Debug)]
struct RejectCommandFailedToExecute;
impl warp::reject::Reject for RejectCommandFailedToExecute {}
#[derive(Debug)]
struct RejectBadRequest;
impl warp::reject::Reject for RejectBadRequest {}

#[derive(Debug)]
struct CommandError {
    stdout: String,
    stderr: String,
}
impl warp::reject::Reject for CommandError {}
impl From<std::process::Output> for CommandError {
    fn from(value: std::process::Output) -> Self {
        CommandError {
            stdout: std::str::from_utf8(value.stdout.as_ref()).unwrap().into(),
            stderr: std::str::from_utf8(value.stderr.as_ref()).unwrap().into(),
        }
    }
}

fn split_allowed_ips_text_line(text: &str) -> (WgPeer, Vec<cidr::IpInet>) {
    let mut string_parts = text.split(|c| c == '\t' || c == ' ');
    let peer = string_parts.next().unwrap();
    let cidrs = string_parts
        .map(|p| cidr::IpInet::from_str(p).unwrap())
        .collect();
    (peer.into(), cidrs)
}
fn split_allowed_ips_text(text: &str, peer: WgPeer) -> (WgPeer, Vec<cidr::IpInet>) {
    text.split('\n')
        .map(|p| split_allowed_ips_text_line(p))
        .filter(|l| l.0 == peer)
        .next()
        .unwrap()
}

async fn wg_modify_address(
    config: wg::Wg,
    message: model::IpcMessage,
) -> Result<impl warp::Reply, warp::Rejection> {
    let peer = message.peer;
    let event = message.event;
    log::debug!("Received event {:?} from peer {:?}", &event, &peer);
    let mut new_peer: Option<&wg::Peer> = None;
    for peer_ in config.peers.iter() {
        if peer_.public_key == peer {
            new_peer = Some(peer_);
            break;
        }
    }
    if new_peer.is_none() {
        return Err(warp::reject::custom(RejectBadRequest));
    }
    let new_peer = new_peer.unwrap();
    let iter = new_peer.allowed_ips.iter();
    let e_info = match event.clone() {
        event::Event::Add(e) => e,
        event::Event::Del(e) => e,
    };
    let iter = match event {
        event::Event::Add(_) => iter.chain(std::iter::once(&e_info.cidr)).collect::<Vec<_>>(),
        _ => iter.collect::<Vec<_>>(),
    };
    let ips_str =
        iter.iter()
            .map(|x| x.to_string())
            .fold("".to_string(), |acc, x| {
                let mut s = String::from(acc);
                s.push_str(&x);
                s.push_str(",");
                s
            });
    let ips_str = &ips_str[0..ips_str.len() - 1];

    let mut command = tokio::process::Command::new("wg");
    command
        .arg("set")
        .arg(&e_info.link)
        .arg("peer")
        .arg(&peer)
        .arg("allowed-ips")
        .arg(ips_str);
    let output = match command_output(&mut command).await {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };

    if output.status.success() {
        Ok("")
    } else {
        Err(warp::reject::custom(CommandError::from(output)))
    }
}

async fn wg_add_address(
    link: WgLink,
    peer: WgPeer,
    cidr: cidr::IpInet,
) -> Result<impl warp::Reply, warp::Rejection> {
    let output = match command_output(
        tokio::process::Command::new("wg")
            .arg("show")
            .arg(&link)
            .arg("allowed-ips"),
    )
    .await
    {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };
    if !output.status.success() {
        return Err(warp::reject::custom(CommandError::from(output)));
    }
    let (peer, ips) =
        split_allowed_ips_text(std::str::from_utf8(output.stdout.as_ref()).unwrap(), peer);
    let ips_str =
        ips.iter()
            .chain([cidr].iter())
            .map(|x| x.to_string())
            .fold("".to_string(), |acc, x| {
                let mut s = String::from(acc);
                s.push_str(&x);
                s.push_str(",");
                s
            });
    let ips_str = &ips_str[0..ips_str.len() - 1];
    let mut command = tokio::process::Command::new("wg");
    command
        .arg("set")
        .arg(&link)
        .arg("peer")
        .arg(&peer)
        .arg("allowed-ips")
        .arg(ips_str);
    let output = match command_output(&mut command).await {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };

    if output.status.success() {
        Ok("")
    } else {
        Err(warp::reject::custom(CommandError::from(output)))
    }
}
async fn wg_del_address(
    link: WgLink,
    peer: WgPeer,
    cidr: cidr::IpInet,
) -> Result<impl warp::Reply, warp::Rejection> {
    let output = match tokio::process::Command::new("wg")
        .arg("show")
        .arg(&link)
        .arg("allowed-ips")
        .output()
        .await
    {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };
    if !output.status.success() {
        return Err(warp::reject::custom(CommandError::from(output)));
    }
    let (peer, ips) =
        split_allowed_ips_text(std::str::from_utf8(output.stdout.as_ref()).unwrap(), peer);
    let ips_str = ips
        .iter()
        .filter(|x| **x != cidr)
        .map(|x| x.to_string())
        .fold("".to_string(), |acc, x| {
            let mut s = String::from(acc);
            s.push_str(&x);
            s.push_str(",");
            s
        });
    let ips_str = &ips_str[0..ips_str.len() - 1];
    let mut command = tokio::process::Command::new("wg");
    command
        .arg("set")
        .arg(&link)
        .arg("peer")
        .arg(&peer)
        .arg("allowed-ips")
        .arg(ips_str);
    println!("command = {:?}", command);
    let output = match command.output().await {
        Ok(v) => v,
        Err(_) => return Err(warp::reject::custom(RejectCommandFailedToExecute)),
    };
    println!("{:?}", output);

    if output.status.success() {
        Ok("")
    } else {
        Err(warp::reject::custom(CommandError::from(output)))
    }
}

async fn send_daemon_get_public_key(iface: &WgLink) -> WgPeer {
    let mut command = tokio::process::Command::new("wg");
    command.arg("show").arg(&iface).arg("public-key");
    let output = helpers::command_output(&mut command).await.unwrap();
    std::str::from_utf8(output.stdout.as_ref())
        .unwrap()
        .trim_end_matches("\n")
        .to_string()
}

async fn event_generator(iface: WgLink, tx: tokio::sync::broadcast::Sender<event::Event>) -> () {
    let mut command = tokio::process::Command::new("ip");
    command.arg("monitor").arg("address").arg("dev").arg(&iface);
    command.stdout(std::process::Stdio::piped());
    command.stderr(std::process::Stdio::piped());
    log::debug!("Shell command {:?}", command);
    let mut child = command.spawn().unwrap();
    let stdout = child.stdout.take().unwrap();
    let stderr = child.stderr.take().unwrap();
    let mut reader = BufReader::new(stdout);
    let mut stderr_reader = BufReader::new(stderr);

    let mut buffer: Vec<u8> = vec![];
    let mut stderr_buffer: Vec<u8> = vec![];
    let mut enable_stdout = true;
    let mut enable_stderr = true;
    loop {
        let b: String;

        if enable_stdout & enable_stderr {
            tokio::select!(
                v = reader.read_until('\n' as u8, &mut buffer) => {
                    if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                        enable_stdout = false;
                        continue;
                    }
                    b = String::from_utf8(buffer).unwrap();
                    buffer = vec![];
                },
                v = stderr_reader.read_until('\n' as u8, &mut stderr_buffer) => {
                    if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                        enable_stderr = false;
                        continue;
                    }
                    let b = String::from_utf8(stderr_buffer).unwrap();
                    stderr_buffer = vec![];
                    log::error!("Monitor produced error: {:?}", &b);
                    continue;
                },
            );
        } else if enable_stdout {
            let v = reader.read_until('\n' as u8, &mut buffer).await;
            if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                enable_stdout = false;
                continue;
            }
            b = String::from_utf8(buffer).unwrap();
            buffer = vec![];
        } else if enable_stderr {
            let v = stderr_reader
                .read_until('\n' as u8, &mut stderr_buffer)
                .await;
            if v.is_err() | v.is_ok_and(|v_| v_ == 0) {
                enable_stderr = false;
                continue;
            }
            let b = String::from_utf8(stderr_buffer).unwrap();
            stderr_buffer = vec![];
            log::error!("Monitor produced error: {:?}", &b);
            continue;
        } else {
            break;
        }

        let e = match event::Event::from_str(&b) {
            Ok(e) => e,
            Err(_) => { continue; },
        };
        tx.send(e).unwrap();
    }
}

async fn event_sender(
    config: wg::Wg,
    addr: std::net::SocketAddr,
    rx: tokio::sync::broadcast::Receiver<event::Event>,
) -> () {
    let mut delayed: HashMap<wg::Peer, Vec<event::Event>> = HashMap::new();
    let mut rx = rx;
    for peer in config.peers.iter() {
        delayed.insert(peer.clone(), vec![]);
    }
    let port = addr.port();
    let client = hyper::Client::new();
    loop {
        let e: event::Event = rx.recv().await.unwrap();
        let public_key = send_daemon_get_public_key(&e.inner().link).await;

        let mut join_set = tokio::task::JoinSet::new();
        join_set.spawn(async {tokio::time::sleep(std::time::Duration::from_millis(500)).await; 0});
        for (i, peer) in config.peers.iter().enumerate() {
            //let u = format!(
            //    "http://{peer}:{port}/wireguard/{iface}/peer/{public_key}/address",
            //    peer=peer.allowed_ips.get(0).unwrap(),
            //    port=port,
            //    iface=encoded_iface,
            //    public_key=encoded_public_key,
            //);
            //let request = hyper::Request::builder()
            //    .method(&method)
            //    .uri(&u)
            //    .body(hyper::Body::from(encoded_cidr.to_string()))
            //    .unwrap();
            let u = format!(
                "http://{peer}:{port}/ipc",
                peer=peer.allowed_ips.get(0).unwrap(),
                port=port,
            );
            let request = hyper::Request::builder()
                .method(&hyper::Method::PATCH)
                .uri(&u)
                .body(hyper::Body::from(serde_json::to_string(&model::IpcMessage {
                    peer: public_key.clone(),
                    event: e.clone(),
                }).unwrap()))
                .unwrap();
            let a_client = client.clone();
            join_set.spawn(async move {
                log::debug!("Request: {:?}", &request);
                let r = a_client.request(request).await;
                log::debug!("Response: {:?}", &r);
                i+1
            });
        }
        while let Some(joined) = join_set.join_next().await {
            let idx = joined.unwrap();
            if idx == 0 {
                log::debug!("Failed to send update to all peers, aborting remaining tasks");
                break;
            }
        }
        join_set.shutdown().await;
    }
}

const BUILD_INFO: &str = include!(concat!(env!("OUT_DIR"), "/buildinfo.txt"));

#[tokio::main]
async fn main() {
    if std::env::var("WGVIRTIPD_LOG").is_err() {
        std::env::set_var("WGVIRTIPD_LOG", "wgvirtipd=info");
    }
    env_logger::init_from_env("WGVIRTIPD_LOG");
    let args = CliArguments::parse();
    let config = wg::Wg::from_file(&args.link).await;

    let alive = warp::path!("alive").and(warp::get()).map(|| BUILD_INFO);
    let closure_config = config.clone();
    let ipc = warp::path!("ipc")
        .and(warp::patch())
        .map(move || closure_config.clone())
        .and(warp::body::json::<model::IpcMessage>())
        .and_then(wg_modify_address);
    let routes = alive.or(ipc);

    log::info!("{}", BUILD_INFO);
    let (tx, rx1) = tokio::sync::broadcast::channel(16);

    tokio::select!(
        _ = warp::serve(routes).run(args.addr) => {},
        _ = event_generator(args.link, tx) => {},
        _ = event_sender(config, args.addr, rx1) => {},
    );
}
